import inspect
import re
import typing as t

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import pytest
import sphinx_autodoc_typehints as sat
from sphinx.application import Sphinx

from scanpydoc.elegant_typehints.formatting import (
    format_annotation,
    _format_terse,
    _format_full,
)
from scanpydoc.elegant_typehints.return_tuple import process_docstring


TestCls = type("Class", (), {})
TestCls.__module__ = "_testmod"


@pytest.fixture
def app(make_app_setup) -> Sphinx:
    return make_app_setup(
        extensions=[
            "sphinx.ext.napoleon",
            "sphinx_autodoc_typehints",
            "scanpydoc.elegant_typehints",
        ],
        qualname_overrides={"_testmod.Class": "test.Class"},
    )


@pytest.fixture
def process_doc(app):
    app.config.typehints_fully_qualified = True

    def process(fn: t.Callable) -> t.List[str]:
        lines = inspect.getdoc(fn).split("\n")
        sat.process_docstring(app, "function", fn.__name__, fn, None, lines)
        process_docstring(app, "function", fn.__name__, fn, None, lines)
        return lines

    return process


def test_app(app):
    assert "qualname_overrides" in app.config.values
    assert "_testmod.Class" in app.config.qualname_overrides


def test_default(app):
    assert format_annotation(str) == ":py:class:`str`"


def test_alternatives(process_doc):
    def fn_test(s: str):
        """
        :param s: Test
        """

    assert process_doc(fn_test) == [
        ":type s: "
        r":annotation-terse:`:py:class:\`str\``\ "
        r":annotation-full:`:py:class:\`str\``",
        ":param s: Test",
    ]


def test_defaults_simple(process_doc):
    def fn_test(s: str = "foo", n: None = None, i: int = 1):
        """
        :param s: Test S
        :param n: Test N
        :param i: Test I
        """

    assert process_doc(fn_test) == [
        ":type s: "
        r":annotation-terse:`:py:class:\`str\``\ "
        r":annotation-full:`:py:class:\`str\`` "
        "(default: ``'foo'``)",
        ":param s: Test S",
        ":type n: "
        r":annotation-terse:`\`\`None\`\``\ "
        r":annotation-full:`\`\`None\`\`` "
        "(default: ``None``)",
        ":param n: Test N",
        ":type i: "
        r":annotation-terse:`:py:class:\`int\``\ "
        r":annotation-full:`:py:class:\`int\`` "
        "(default: ``1``)",
        ":param i: Test I",
    ]


def test_defaults_complex(process_doc):
    def fn_test(m: t.Mapping[str, int] = {}):
        """
        :param m: Test M
        """

    assert process_doc(fn_test) == [
        ":type m: "
        r":annotation-terse:`:py:class:\`typing.Mapping\``\ "
        r":annotation-full:`"
        r":py:class:\`typing.Mapping\`\[:py:class:\`str\`, :py:class:\`int\`]"
        "` (default: ``{}``)",
        ":param m: Test M",
    ]


def test_mapping(app):
    assert _format_terse(t.Mapping[str, t.Any]) == ":py:class:`~typing.Mapping`"
    assert _format_full(t.Mapping[str, t.Any]) == (
        r":py:class:`~typing.Mapping`\["
        r":py:class:`str`, "
        r":py:data:`~typing.Any`"
        r"]"
    )


def test_dict(app):
    assert _format_terse(t.Dict[str, t.Any]) == (
        "{:py:class:`str`: :py:data:`~typing.Any`}"
    )


def test_literal(app):
    assert _format_terse(Literal["str", 1, None]) == "{'str', 1, None}"
    assert _format_full(Literal["str", 1, None]) == (
        r":py:data:`~typing.Literal`\['str', 1, None]"
    )


def test_qualname_overrides(app):
    assert TestCls.__module__ == "_testmod"
    assert _format_terse(TestCls) == ":py:class:`~test.Class`"


def test_qualname_overrides_recursive(app):
    assert _format_terse(t.Union[TestCls, str]) == (
        r":py:class:`~test.Class`, :py:class:`str`"
    )
    assert _format_full(t.Union[TestCls, str]) == (
        r":py:data:`~typing.Union`\["
        r":py:class:`~test.Class`, "
        r":py:class:`str`"
        r"]"
    )


def test_fully_qualified(app):
    assert _format_terse(t.Union[TestCls, str], True) == (
        r":py:class:`test.Class`, :py:class:`str`"
    )
    assert _format_full(t.Union[TestCls, str], True) == (
        r":py:data:`typing.Union`\[" r":py:class:`test.Class`, " r":py:class:`str`" r"]"
    )


def test_classes_get_added(app, parse):
    doc = parse(app, r":annotation-full:`:py:class:\`str\``")
    assert doc[0].tagname == "paragraph"
    assert doc[0][0].tagname == "inline"
    assert doc[0][0]["classes"] == ["annotation", "full"]
    # print(doc.asdom().toprettyxml())


@pytest.mark.parametrize("formatter", [_format_terse, _format_full], ids="tf")
# These guys aren’t listed as classes in Python’s intersphinx index:
@pytest.mark.parametrize(
    "annotation",
    [
        t.Any,
        t.AnyStr,
        # t.NoReturn,
        t.Callable[[int], None],
        # t.ClassVar[t.Any],
        t.Optional[int],
        t.Tuple[int, str],
        t.Tuple[float, ...],
        t.Union[int, str],
    ],
    ids=lambda p: str(p).replace("typing.", ""),
)
def test_typing_classes(app, annotation, formatter):
    name = (
        getattr(annotation, "_name", None)
        or getattr(annotation, "__name__", None)
        or getattr(getattr(annotation, "__origin__", None), "_name", None)
        # 3.6 _Any and _Union
        or annotation.__class__.__name__[1:]
    )
    if name == "Union":
        if formatter is _format_terse:
            pytest.skip("Tested elsewhere")
        elif len(annotation.__args__) == 2 and type(None) in annotation.__args__:
            name = "Optional"
    assert formatter(annotation, True).startswith(f":py:data:`typing.{name}")


def test_typing_class_nested(app):
    assert _format_full(t.Optional[t.Tuple[int, str]]) == (
        ":py:data:`~typing.Optional`\\["
        ":py:data:`~typing.Tuple`\\[:py:class:`int`, :py:class:`str`]"
        "]"
    )


@pytest.mark.parametrize(
    "docstring",
    [
        """
        :param: x
        :return: foo
                     A foo!
                 bar
                     A bar!
        """,
        """
        :return: foo
                     A foo!
                 bar
                     A bar!
        :param: x
        """,
    ],
    ids=["Last", "First"],
)
@pytest.mark.parametrize(
    "return_ann",
    [t.Tuple[str, int], t.Optional[t.Tuple[str, int]]],
    ids=["Tuple", "Optional[Tuple]"],
)
def test_return(process_doc, docstring, return_ann):
    def fn_test():
        pass

    fn_test.__doc__ = docstring
    fn_test.__annotations__["return"] = return_ann
    lines = [
        l
        for l in process_doc(fn_test)
        if not re.match("^:(rtype|param|annotation-(full|terse)):", l)
    ]
    assert lines == [
        r":return: foo : "
        r":annotation-terse:`:py:class:\`str\``\ "
        r":annotation-full:`:py:class:\`str\``",
        "             A foo!",
        r"         bar : "
        r":annotation-terse:`:py:class:\`int\``\ "
        r":annotation-full:`:py:class:\`int\``",
        "             A bar!",
    ]


def test_return_too_many(process_doc):
    def fn_test() -> t.Tuple[int, str]:
        """
        :return: foo
                     A foo!
                 bar
                     A bar!
                 baz
                     A baz!
        """

    assert not any(
        "annotation-terse" in l
        for l in process_doc(fn_test)
        if not l.startswith(":rtype:")
    )



import inspect
from collections import abc
from typing import Any, Union  # Meta
from typing import Type, Mapping, Sequence, Iterable  # ABC
from typing import Dict, List, Tuple  # Concrete

from scanpydoc import elegant_typehints

try:
    from typing import Literal
except ImportError:
    try:
        from typing_extensions import Literal
    except ImportError:
        Literal = object()

import sphinx_autodoc_typehints
from sphinx_autodoc_typehints import format_annotation as _format_orig
from docutils import nodes
from docutils.nodes import Node
from docutils.parsers.rst.roles import set_classes
from docutils.parsers.rst.states import Inliner, Struct
from docutils.utils import SystemMessage, unescape


def _format_full(annotation: Type[Any], fully_qualified: bool = False):
    if inspect.isclass(annotation) and annotation.__module__ == "builtins":
        return _format_orig(annotation, fully_qualified)

    origin = getattr(annotation, "__origin__", None)
    tilde = "" if fully_qualified else "~"

    annotation_cls = annotation if inspect.isclass(annotation) else type(annotation)
    if annotation_cls.__module__ == "typing":
        return _format_orig(annotation, fully_qualified)

    # Only if this is a real class we override sphinx_autodoc_typehints
    if inspect.isclass(annotation) or inspect.isclass(origin):
        full_name = f"{annotation.__module__}.{annotation.__qualname__}"
        override = elegant_typehints.qualname_overrides.get(full_name)
        if override is not None:
            return f":py:class:`{tilde}{override}`"

    return _format_orig(annotation, fully_qualified)


def _format_terse(annotation: Type[Any], fully_qualified: bool = False) -> str:
    origin = getattr(annotation, "__origin__", None)

    union_params = getattr(annotation, "__union_params__", None)
    # display `Union[A, B]` as `A, B`
    if origin is Union or union_params:
        params = union_params or getattr(annotation, "__args__", None)
        # Never use the `Optional` keyword in the displayed docs.
        # Use the more verbose `, None` instead,
        # as is the convention in the other large numerical packages
        # if len(params or []) == 2 and getattr(params[1], '__qualname__', None) == 'NoneType':
        #     return fa_orig(annotation)  # Optional[...]
        return ", ".join(_format_terse(p, fully_qualified) for p in params)

    # do not show the arguments of Mapping
    if origin in (abc.Mapping, Mapping):
        return f":py:class:`{'' if fully_qualified else '~'}typing.Mapping`"

    # display dict as {k: v}
    if origin in (dict, Dict):
        k, v = annotation.__args__
        return f"{{{_format_terse(k, fully_qualified)}: {_format_terse(v, fully_qualified)}}}"

    if origin is Literal or hasattr(annotation, "__values__"):
        values = getattr(annotation, "__args__", ()) or annotation.__values__
        return f"{{{', '.join(map(repr, values))}}}"

    return _format_full(annotation, fully_qualified)


def format_annotation(annotation: Type[Any], fully_qualified: bool = False) -> str:
    r"""Generate reStructuredText containing links to the types.

    Unlike :func:`sphinx_autodoc_typehints.format_annotation`,
    it tries to achieve a simpler style as seen in numeric packages like numpy.

    Args:
        annotation: A type or class used as type annotation.
        fully_qualified: If links should be formatted as fully qualified
            (e.g. ``:py:class:`foo.Bar```) or not (e.g. ``:py:class:`~foo.Bar```).

    Returns:
        reStructuredText describing the type
    """
    if sphinx_autodoc_typehints.format_annotation is not format_annotation:
        raise RuntimeError(
            "This function is not guaranteed to work correctly without overriding"
            "`sphinx_autodoc_typehints.format_annotation` with it."
        )

    curframe = inspect.currentframe()
    calframe = inspect.getouterframes(curframe, 2)
    if calframe[1][3] == "process_docstring":
        annot_fmt = format_both(annotation, fully_qualified)
        if elegant_typehints.annotate_defaults:
            variables = calframe[1].frame.f_locals
            sig = inspect.signature(variables["obj"])
            if variables["argname"] != "return":
                default = sig.parameters[variables["argname"]].default
                if default is not inspect.Parameter.empty:
                    annot_fmt += f" (default: ``{_escape(repr(default))}``)"
        return annot_fmt
    else:  # recursive use
        return _format_full(annotation, fully_qualified)


def format_both(annotation: Type[Any], fully_qualified: bool = False):
    return (
        f":annotation-terse:`{_escape(_format_terse(annotation, fully_qualified))}`\\ "
        f":annotation-full:`{_escape(_format_full(annotation, fully_qualified))}`"
    )


def _role_annot(
    name: str,
    rawtext: str,
    text: str,
    lineno: int,
    inliner: Inliner,
    options: Dict[str, Any] = {},
    content: Sequence[str] = (),
    # *,  # https://github.com/ambv/black/issues/613
    additional_classes: Iterable[str] = (),
) -> Tuple[List[Node], List[SystemMessage]]:
    options = options.copy()
    set_classes(options)
    if additional_classes:
        options["classes"] = options.get("classes", []).copy()
        options["classes"].extend(additional_classes)
    memo = Struct(
        document=inliner.document, reporter=inliner.reporter, language=inliner.language
    )
    node = nodes.inline(unescape(rawtext), "", **options)
    children, messages = inliner.parse(_unescape(text), lineno, memo, node)
    node.extend(children)
    return [node], messages


def _escape(rst: str) -> str:
    return rst.replace("`", "\\`")


def _unescape(rst: str) -> str:
    # TODO: IDK why the [ part is necessary.
    return unescape(rst).replace("\\`", "`").replace("[", "\\[")
